@@ -31,6 +31,7 @@
 #include <glib/gthread.h>
 #include <glib/gslice.h>
 
+#include "gtestutils.h"
 #include "gthreadprivate.h"
 
 #ifdef G_BIT_LOCK_FORCE_FUTEX_EMULATION
@@ -55,7 +56,7 @@ static GSList *g_futex_address_list = NULL;
 
 /* < private >
  * g_futex_wait:
- * @address: a pointer to an integer
+ * @address: (type gpointer): a pointer to an integer
  * @value: the value that should be at @address
  *
  * Atomically checks that the value stored at @address is equal to
@@ -79,7 +80,7 @@ g_futex_wait (const gint *address,
 
 /* < private >
  * g_futex_wake:
- * @address: a pointer to an integer
+ * @address: (type gpointer): a pointer to an integer
  *
  * Nominally, wakes one thread that is blocked in g_futex_wait() on
  * @address (if any thread is currently waiting).
@@ -175,7 +176,7 @@ g_futex_wake (const gint *address)
 static gint g_bit_lock_contended[CONTENTION_CLASSES];  /* (atomic) */
 
 G_ALWAYS_INLINE static inline guint
-bit_lock_contended_class (gpointer address)
+bit_lock_contended_class (gconstpointer address)
 {
   return ((gsize) address) % G_N_ELEMENTS (g_bit_lock_contended);
 }
@@ -186,9 +187,103 @@ bit_lock_contended_class (gpointer address)
   #endif
 #endif
 
+static const gint *g_futex_int_address (const void *address);
+
+G_ALWAYS_INLINE static inline void
+bit_lock_futex_wait (gconstpointer address, gboolean is_pointer_pointer, gint value)
+{
+  const guint CLASS = bit_lock_contended_class (address);
+
+  g_atomic_int_add (&g_bit_lock_contended[CLASS], +1);
+  if (is_pointer_pointer)
+    address = g_futex_int_address (address);
+  g_futex_wait (address, value);
+  g_atomic_int_add (&g_bit_lock_contended[CLASS], -1);
+}
+
+G_ALWAYS_INLINE static inline void
+bit_lock_futex_maybe_wake (gconstpointer address, gboolean is_pointer_pointer)
+{
+  const guint CLASS = bit_lock_contended_class (address);
+
+  /* Warning: unlocking may allow another thread to proceed and destroy the
+   * memory that @address points to. We thus must not dereference it anymore.
+   */
+
+  if (g_atomic_int_get (&g_bit_lock_contended[CLASS]))
+    {
+      if (is_pointer_pointer)
+        address = g_futex_int_address (address);
+      g_futex_wake (address);
+    }
+}
+
+/**
+ * g_bit_lock_and_get:
+ * @address: (type gpointer): a pointer to an integer
+ * @lock_bit: a bit value between 0 and 31
+ * @out_val: (out) (optional): return location for the new value of the integer
+ *
+ * Sets the indicated @lock_bit in @address and atomically returns the new value.
+ *
+ * This is like [func@GLib.bit_lock], except it can atomically return the new value at
+ * @address (right after obtaining the lock). Thus the value returned in @out_val
+ * always has the @lock_bit set.
+ *
+ * Since: 2.86
+ **/
+void
+g_bit_lock_and_get (gint *address,
+                    guint lock_bit,
+                    gint *out_val)
+{
+  const guint MASK = 1u << lock_bit;
+  guint v;
+
+#ifdef G_ENABLE_DEBUG
+  g_assert (lock_bit < 32u);
+#endif
+
+#ifdef USE_ASM_GOTO
+  if (G_LIKELY (!out_val))
+    {
+      while (TRUE)
+        {
+          __asm__ volatile goto("lock bts %1, (%0)\n"
+                                "jc %l[contended]"
+                                : /* no output */
+                                : "r"(address), "r"(lock_bit)
+                                : "cc", "memory"
+                                : contended);
+          return;
+
+        contended:
+          {
+            guint v;
+
+            v = (guint) g_atomic_int_get (address);
+            if (v & MASK)
+              bit_lock_futex_wait (address, FALSE, (gint) v);
+          }
+        }
+    }
+#endif
+
+retry:
+  v = g_atomic_int_or ((guint *) address, MASK);
+  if (v & MASK)
+    {
+      bit_lock_futex_wait (address, FALSE, (gint) v);
+      goto retry;
+    }
+
+  if (out_val)
+    *out_val = (gint) (v | MASK);
+}
+
 /**
  * g_bit_lock:
- * @address: a pointer to an integer
+ * @address: (type gpointer): a pointer to an integer
  * @lock_bit: a bit value between 0 and 31
  *
  * Sets the indicated @lock_bit in @address.  If the bit is already
@@ -212,57 +307,12 @@ void
 g_bit_lock (volatile gint *address,
             gint           lock_bit)
 {
-  gint *address_nonvolatile = (gint *) address;
-
-#ifdef USE_ASM_GOTO
- retry:
-  __asm__ volatile goto ("lock bts %1, (%0)\n"
-                         "jc %l[contended]"
-                         : /* no output */
-                         : "r" (address), "r" (lock_bit)
-                         : "cc", "memory"
-                         : contended);
-  return;
-
- contended:
-  {
-    guint mask = 1u << lock_bit;
-    guint v;
-
-    v = (guint) g_atomic_int_get (address_nonvolatile);
-    if (v & mask)
-      {
-        guint class = bit_lock_contended_class (address_nonvolatile);
-
-        g_atomic_int_add (&g_bit_lock_contended[class], +1);
-        g_futex_wait (address_nonvolatile, v);
-        g_atomic_int_add (&g_bit_lock_contended[class], -1);
-      }
-  }
-  goto retry;
-#else
-  guint mask = 1u << lock_bit;
-  guint v;
-
- retry:
-  v = g_atomic_int_or (address_nonvolatile, mask);
-  if (v & mask)
-    /* already locked */
-    {
-      guint class = bit_lock_contended_class (address_nonvolatile);
-
-      g_atomic_int_add (&g_bit_lock_contended[class], +1);
-      g_futex_wait (address_nonvolatile, v);
-      g_atomic_int_add (&g_bit_lock_contended[class], -1);
-
-      goto retry;
-    }
-#endif
+  g_bit_lock_and_get ((gint *) address, (guint) lock_bit, NULL);
 }
 
 /**
  * g_bit_trylock:
- * @address: a pointer to an integer
+ * @address: (type gpointer) a pointer to an integer
  * @lock_bit: a bit value between 0 and 31
  *
  * Sets the indicated @lock_bit in @address, returning %TRUE if
@@ -311,7 +361,7 @@ g_bit_trylock (volatile gint *address,
 
 /**
  * g_bit_unlock:
- * @address: a pointer to an integer
+ * @address: (type gpointer): a pointer to an integer
  * @lock_bit: a bit value between 0 and 31
  *
  * Clears the indicated @lock_bit in @address.  If another thread is
@@ -345,15 +395,62 @@ g_bit_unlock (volatile gint *address,
   /* Warning: unlocking may allow another thread to proceed and destroy the
    * memory that @address points to. We thus must not dereference it anymore.
    */
-
-  {
-    guint class = bit_lock_contended_class (address_nonvolatile);
-
-    if (g_atomic_int_get (&g_bit_lock_contended[class]))
-      g_futex_wake (address_nonvolatile);
-  }
+  bit_lock_futex_maybe_wake (address_nonvolatile, FALSE);
 }
 
+/**
+ * g_bit_unlock_and_set:
+ * @address: (type gpointer): a pointer to an integer
+ * @lock_bit: a bit value between 0 and 31
+ * @val: the new value to set
+ * @preserve_mask: mask for bits from @address to preserve
+ *
+ * This is like [func@GLib.bit_unlock]() but also atomically sets @address to
+ * @val.
+ *
+ * If @preserve_mask is not zero, then the @preserve_mask bits will be
+ * preserved in @address and are not set to @val.
+ *
+ * Note that the @lock_bit bit will be always unset regardless of
+ * @val, @preserve_mask and the currently set value in @address.
+ *
+ * Since: 2.86
+ **/
+void
+g_bit_unlock_and_set (gint *address,
+                      guint lock_bit,
+                      gint val,
+                      gint preserve_mask)
+
+{
+  const guint MASK = 1u << lock_bit;
+
+#ifdef G_ENABLE_DEBUG
+  g_assert (lock_bit < 32u);
+#endif
+
+  if (G_UNLIKELY (preserve_mask != 0))
+    {
+      guint old_val;
+      guint new_val;
+
+      old_val = (guint) g_atomic_int_get (address);
+
+    again:
+      new_val = ((old_val & ((guint) preserve_mask)) | (((guint) val) & ~((guint) preserve_mask))) & ~MASK;
+      if (!g_atomic_int_compare_and_exchange_full (address, (gint) old_val, (gint) new_val, (gint *) &old_val))
+        goto again;
+    }
+  else
+    {
+      g_atomic_int_set (address, (gint) (((guint) val) & ~MASK));
+    }
+
+  /* Warning: unlocking may allow another thread to proceed and destroy the
+   * memory that @address points to. We thus must not dereference it anymore.
+   */
+  bit_lock_futex_maybe_wake (address, FALSE);
+}
 
 /* We emulate pointer-sized futex(2) because the kernel API only
  * supports integers.
@@ -379,7 +476,7 @@ g_bit_unlock (volatile gint *address,
  *
  *   g_futex_wake (g_futex_int_address (int_address));
  */
-static const gint *
+G_ALWAYS_INLINE static inline const gint *
 g_futex_int_address (const void *address)
 {
   const gint *int_address = address;
@@ -444,7 +541,6 @@ void
                               guint lock_bit,
                               guintptr *out_ptr)
 {
-  guint class = bit_lock_contended_class (address);
   guintptr mask;
   guintptr v;
 
@@ -468,11 +564,7 @@ void
         contended:
           v = (guintptr) g_atomic_pointer_get ((gpointer *) address);
           if (v & mask)
-            {
-              g_atomic_int_add (&g_bit_lock_contended[class], +1);
-              g_futex_wait (g_futex_int_address (address), v);
-              g_atomic_int_add (&g_bit_lock_contended[class], -1);
-            }
+            bit_lock_futex_wait (address, TRUE, (gint) v);
         }
     }
 #endif
@@ -480,11 +572,8 @@ void
 retry:
   v = g_atomic_pointer_or ((gpointer *) address, mask);
   if (v & mask)
-    /* already locked */
     {
-      g_atomic_int_add (&g_bit_lock_contended[class], +1);
-      g_futex_wait (g_futex_int_address (address), (guint) v);
-      g_atomic_int_add (&g_bit_lock_contended[class], -1);
+      bit_lock_futex_wait (address, TRUE, (gint) v);
       goto retry;
     }
 
@@ -606,13 +695,7 @@ void
     /* Warning: unlocking may allow another thread to proceed and destroy the
      * memory that @address points to. We thus must not dereference it anymore.
      */
-
-    {
-      guint class = bit_lock_contended_class (address_nonvolatile);
-
-      if (g_atomic_int_get (&g_bit_lock_contended[class]))
-        g_futex_wake (g_futex_int_address (address_nonvolatile));
-    }
+    bit_lock_futex_maybe_wake (address_nonvolatile, TRUE);
   }
 }
 
@@ -653,8 +736,8 @@ g_pointer_bit_lock_mask_ptr (gpointer ptr, guint lock_bit, gboolean set, guintpt
  * @ptr: the new pointer value to set
  * @preserve_mask: if non-zero, those bits of the current pointer in @address
  *   are preserved.
- *   Note that the @lock_bit bit will be always set according to @set,
- *   regardless of @preserve_mask and the currently set value in @address.
+ *   Note that the @lock_bit bit will be always unset regardless of
+ *   @ptr, @preserve_mask and the currently set value in @address.
  *
  * This is equivalent to g_pointer_bit_unlock() and atomically setting
  * the pointer value.
@@ -672,7 +755,6 @@ void (g_pointer_bit_unlock_and_set) (void *address,
                                      guintptr preserve_mask)
 {
   gpointer *pointer_address = address;
-  guint class = bit_lock_contended_class (address);
   gpointer ptr2;
 
   g_return_if_fail (lock_bit < 32u);
@@ -692,8 +774,7 @@ void (g_pointer_bit_unlock_and_set) (void *address,
       g_atomic_pointer_set (pointer_address, ptr2);
     }
 
-  if (g_atomic_int_get (&g_bit_lock_contended[class]) > 0)
-    g_futex_wake (g_futex_int_address (address));
+  bit_lock_futex_maybe_wake (address, TRUE);
 
   /* It makes no sense, if unlocking mangles the pointer. Assert against
    * that.
