@@ -153,7 +153,7 @@ test_default_handler_bar_info (void)
   g_log_writer_default_set_use_stderr (FALSE);
   g_log_set_default_handler (g_log_default_handler, NULL);
 
-  g_setenv ("G_MESSAGES_DEBUG", "foo bar baz", TRUE);
+  g_assert_cmpstr (g_getenv ("G_MESSAGES_DEBUG"), ==, "foo bar baz");
 
   g_log ("bar", G_LOG_LEVEL_INFO, "message5");
   exit (0);
@@ -165,7 +165,7 @@ test_default_handler_baz_debug (void)
   g_log_writer_default_set_use_stderr (FALSE);
   g_log_set_default_handler (g_log_default_handler, NULL);
 
-  g_setenv ("G_MESSAGES_DEBUG", "foo bar baz", TRUE);
+  g_assert_cmpstr (g_getenv ("G_MESSAGES_DEBUG"), ==, "foo bar baz");
 
   g_log ("baz", G_LOG_LEVEL_DEBUG, "message6");
   exit (0);
@@ -177,7 +177,7 @@ test_default_handler_debug (void)
   g_log_writer_default_set_use_stderr (FALSE);
   g_log_set_default_handler (g_log_default_handler, NULL);
 
-  g_setenv ("G_MESSAGES_DEBUG", "all", TRUE);
+  g_assert_cmpstr (g_getenv ("G_MESSAGES_DEBUG"), ==, "all");
 
   g_log ("foo", G_LOG_LEVEL_DEBUG, "6");
   g_log ("bar", G_LOG_LEVEL_DEBUG, "6");
@@ -192,7 +192,7 @@ test_default_handler_debug_stderr (void)
   g_log_writer_default_set_use_stderr (TRUE);
   g_log_set_default_handler (g_log_default_handler, NULL);
 
-  g_setenv ("G_MESSAGES_DEBUG", "all", TRUE);
+  g_assert_cmpstr (g_getenv ("G_MESSAGES_DEBUG"), ==, "all");
 
   g_log ("foo", G_LOG_LEVEL_DEBUG, "6");
   g_log ("bar", G_LOG_LEVEL_DEBUG, "6");
@@ -201,10 +201,19 @@ test_default_handler_debug_stderr (void)
   exit (0);
 }
 
+static void
+test_default_handler_would_drop_env_systemd (void)
+{
+  g_assert_cmpstr (g_getenv ("DEBUG_INVOCATION"), ==, "1");
+
+  g_assert_false (g_log_writer_default_would_drop (G_LOG_LEVEL_DEBUG, "foo"));
+  g_assert_false (g_log_writer_default_would_drop (G_LOG_LEVEL_DEBUG, "bar"));
+}
+
 static void
 test_default_handler_would_drop_env5 (void)
 {
-  g_setenv ("G_MESSAGES_DEBUG", "foobar", TRUE);
+  g_assert_cmpstr (g_getenv ("G_MESSAGES_DEBUG"), ==, "foobar");
 
   g_assert_true (g_log_writer_default_would_drop (G_LOG_LEVEL_DEBUG, "foo"));
   g_assert_true (g_log_writer_default_would_drop (G_LOG_LEVEL_DEBUG, "bar"));
@@ -213,7 +222,7 @@ test_default_handler_would_drop_env5 (void)
 static void
 test_default_handler_would_drop_env4 (void)
 {
-  g_setenv ("G_MESSAGES_DEBUG", "all", TRUE);
+  g_assert_cmpstr (g_getenv ("G_MESSAGES_DEBUG"), ==, "all");
 
   g_assert_false (g_log_writer_default_would_drop (G_LOG_LEVEL_ERROR, "foo"));
   g_assert_false (g_log_writer_default_would_drop (G_LOG_LEVEL_CRITICAL, "foo"));
@@ -227,7 +236,7 @@ test_default_handler_would_drop_env4 (void)
 static void
 test_default_handler_would_drop_env3 (void)
 {
-  g_setenv ("G_MESSAGES_DEBUG", "foo bar", TRUE);
+  g_assert_cmpstr (g_getenv ("G_MESSAGES_DEBUG"), ==, "foo bar");
 
   g_assert_false (g_log_writer_default_would_drop (G_LOG_LEVEL_ERROR, "foo"));
   g_assert_false (g_log_writer_default_would_drop (G_LOG_LEVEL_CRITICAL, "foo"));
@@ -241,7 +250,7 @@ test_default_handler_would_drop_env3 (void)
 static void
 test_default_handler_would_drop_env2 (void)
 {
-  g_setenv ("G_MESSAGES_DEBUG", "  bar    baz ", TRUE);
+  g_assert_cmpstr (g_getenv ("G_MESSAGES_DEBUG"), ==, "  bar    baz ");
 
   g_assert_false (g_log_writer_default_would_drop (G_LOG_LEVEL_ERROR, "foo"));
   g_assert_false (g_log_writer_default_would_drop (G_LOG_LEVEL_CRITICAL, "foo"));
@@ -255,7 +264,7 @@ test_default_handler_would_drop_env2 (void)
 static void
 test_default_handler_would_drop_env1 (void)
 {
-  g_unsetenv ("G_MESSAGES_DEBUG");
+  g_assert_null (g_getenv ("G_MESSAGES_DEBUG"));
 
   g_assert_false (g_log_writer_default_would_drop (G_LOG_LEVEL_ERROR, "foo"));
   g_assert_false (g_log_writer_default_would_drop (G_LOG_LEVEL_CRITICAL, "foo"));
@@ -269,7 +278,7 @@ test_default_handler_would_drop_env1 (void)
 static void
 test_default_handler_would_drop (void)
 {
-  g_unsetenv ("G_MESSAGES_DEBUG");
+  g_assert_null (g_getenv ("G_MESSAGES_DEBUG"));
 
   g_assert_false (g_log_writer_default_would_drop (G_LOG_LEVEL_ERROR, "foo"));
   g_assert_false (g_log_writer_default_would_drop (G_LOG_LEVEL_CRITICAL, "foo"));
@@ -435,106 +444,187 @@ test_default_handler_0x400 (void)
   exit (0);
 }
 
+static void
+test_default_handler_structured_logging_non_nul_terminated_strings (void)
+{
+  g_log_writer_default_set_use_stderr (FALSE);
+  g_log_set_default_handler (g_log_default_handler, NULL);
+  g_assert_cmpstr (g_getenv ("G_MESSAGES_DEBUG"), ==, "foo");
+
+  const gchar domain_1[] = {'f', 'o', 'o' };
+  const gchar domain_2[] = { 'b', 'a', 'r' };
+  const gchar message_1[] = { 'b', 'a', 'z' };
+  const gchar message_2[] = { 'b', 'l', 'a' };
+  const GLogField fields[] = {
+    { "GLIB_DOMAIN", domain_1, sizeof (domain_1) },
+    { "MESSAGE", message_1, sizeof (message_1) },
+  };
+  const GLogField other_fields[] = {
+    { "GLIB_DOMAIN", domain_2, sizeof (domain_2) },
+    { "MESSAGE", message_2, sizeof (message_2) },
+  };
+
+  g_log_structured_array (G_LOG_LEVEL_DEBUG, fields, G_N_ELEMENTS (fields));
+  g_log_structured_array (G_LOG_LEVEL_DEBUG, other_fields, G_N_ELEMENTS (other_fields));
+
+  exit (0);
+}
+
+/* Helper wrapper around g_test_trap_subprocess_with_envp() which sets the
+ * logging-related environment variables. `NULL` will unset a variable. */
+static void
+test_trap_subprocess_with_logging_envp (const char *test_path,
+                                        const char *g_messages_debug,
+                                        const char *debug_invocation)
+{
+  char **envp = g_get_environ ();
+
+  if (g_messages_debug != NULL)
+    envp = g_environ_setenv (g_steal_pointer (&envp), "G_MESSAGES_DEBUG", g_messages_debug, TRUE);
+  else
+    envp = g_environ_unsetenv (g_steal_pointer (&envp), "G_MESSAGES_DEBUG");
+
+  if (debug_invocation != NULL)
+    envp = g_environ_setenv (g_steal_pointer (&envp), "DEBUG_INVOCATION", debug_invocation, TRUE);
+  else
+    envp = g_environ_unsetenv (g_steal_pointer (&envp), "DEBUG_INVOCATION");
+
+  g_test_trap_subprocess_with_envp (test_path, (const char * const *) envp, 0, G_TEST_SUBPROCESS_DEFAULT);
+
+  g_strfreev (envp);
+}
+
 static void
 test_default_handler (void)
 {
-  g_test_trap_subprocess ("/logging/default-handler/subprocess/error", 0,
-                          G_TEST_SUBPROCESS_DEFAULT);
+  test_trap_subprocess_with_logging_envp ("/logging/default-handler/subprocess/error",
+                                          NULL, NULL);
   g_test_trap_assert_failed ();
   g_test_trap_assert_stderr ("*ERROR*message1*");
 
-  g_test_trap_subprocess ("/logging/default-handler/subprocess/error-stderr", 0,
-                          G_TEST_SUBPROCESS_DEFAULT);
+  test_trap_subprocess_with_logging_envp ("/logging/default-handler/subprocess/error-stderr",
+                                          NULL, NULL);
   g_test_trap_assert_failed ();
   g_test_trap_assert_stderr ("*ERROR*message1*");
 
-  g_test_trap_subprocess ("/logging/default-handler/subprocess/critical", 0,
-                          G_TEST_SUBPROCESS_DEFAULT);
+  test_trap_subprocess_with_logging_envp ("/logging/default-handler/subprocess/critical",
+                                          NULL, NULL);
   g_test_trap_assert_failed ();
   g_test_trap_assert_stderr ("*CRITICAL*message2*");
 
-  g_test_trap_subprocess ("/logging/default-handler/subprocess/critical-stderr", 0,
-                          G_TEST_SUBPROCESS_DEFAULT);
+  test_trap_subprocess_with_logging_envp ("/logging/default-handler/subprocess/critical-stderr",
+                                          NULL, NULL);
   g_test_trap_assert_failed ();
   g_test_trap_assert_stderr ("*CRITICAL*message2*");
 
-  g_test_trap_subprocess ("/logging/default-handler/subprocess/warning", 0,
-                          G_TEST_SUBPROCESS_DEFAULT);
+  test_trap_subprocess_with_logging_envp ("/logging/default-handler/subprocess/warning",
+                                          NULL, NULL);
   g_test_trap_assert_failed ();
   g_test_trap_assert_stderr ("*WARNING*message3*");
 
-  g_test_trap_subprocess ("/logging/default-handler/subprocess/warning-stderr", 0,
-                          G_TEST_SUBPROCESS_DEFAULT);
+  test_trap_subprocess_with_logging_envp ("/logging/default-handler/subprocess/warning-stderr",
+                                          NULL, NULL);
   g_test_trap_assert_failed ();
   g_test_trap_assert_stderr ("*WARNING*message3*");
 
-  g_test_trap_subprocess ("/logging/default-handler/subprocess/message", 0,
-                          G_TEST_SUBPROCESS_DEFAULT);
+  test_trap_subprocess_with_logging_envp ("/logging/default-handler/subprocess/message",
+                                          NULL, NULL);
   g_test_trap_assert_passed ();
   g_test_trap_assert_stderr ("*Message*message4*");
 
-  g_test_trap_subprocess ("/logging/default-handler/subprocess/message-stderr", 0,
-                          G_TEST_SUBPROCESS_DEFAULT);
+  test_trap_subprocess_with_logging_envp ("/logging/default-handler/subprocess/message-stderr",
+                                          NULL, NULL);
   g_test_trap_assert_passed ();
   g_test_trap_assert_stderr ("*Message*message4*");
 
-  g_test_trap_subprocess ("/logging/default-handler/subprocess/info", 0,
-                          G_TEST_SUBPROCESS_DEFAULT);
+  test_trap_subprocess_with_logging_envp ("/logging/default-handler/subprocess/info",
+                                          NULL, NULL);
   g_test_trap_assert_passed ();
   g_test_trap_assert_stdout_unmatched ("*INFO*message5*");
 
-  g_test_trap_subprocess ("/logging/default-handler/subprocess/info-stderr", 0,
-                          G_TEST_SUBPROCESS_DEFAULT);
+  test_trap_subprocess_with_logging_envp ("/logging/default-handler/subprocess/info-stderr",
+                                          NULL, NULL);
   g_test_trap_assert_passed ();
   g_test_trap_assert_stderr_unmatched ("*INFO*message5*");
 
-  g_test_trap_subprocess ("/logging/default-handler/subprocess/bar-info", 0,
-                          G_TEST_SUBPROCESS_DEFAULT);
+  test_trap_subprocess_with_logging_envp ("/logging/default-handler/subprocess/bar-info",
+                                          "foo bar baz", NULL);
   g_test_trap_assert_passed ();
   g_test_trap_assert_stdout ("*INFO*message5*");
 
-  g_test_trap_subprocess ("/logging/default-handler/subprocess/baz-debug", 0,
-                          G_TEST_SUBPROCESS_DEFAULT);
+  test_trap_subprocess_with_logging_envp ("/logging/default-handler/subprocess/baz-debug",
+                                          "foo bar baz", NULL);
   g_test_trap_assert_passed ();
   g_test_trap_assert_stdout ("*DEBUG*message6*");
 
-  g_test_trap_subprocess ("/logging/default-handler/subprocess/debug", 0,
-                          G_TEST_SUBPROCESS_DEFAULT);
+  test_trap_subprocess_with_logging_envp ("/logging/default-handler/subprocess/debug",
+                                          "all", NULL);
   g_test_trap_assert_passed ();
   g_test_trap_assert_stdout ("*DEBUG*6*6*6*");
 
-  g_test_trap_subprocess ("/logging/default-handler/subprocess/debug-stderr", 0,
-                          G_TEST_SUBPROCESS_DEFAULT);
+  test_trap_subprocess_with_logging_envp ("/logging/default-handler/subprocess/debug-stderr",
+                                          "all", NULL);
   g_test_trap_assert_passed ();
   g_test_trap_assert_stdout_unmatched ("DEBUG");
   g_test_trap_assert_stderr ("*DEBUG*6*6*6*");
 
-  g_test_trap_subprocess ("/logging/default-handler/subprocess/0x400", 0,
-                          G_TEST_SUBPROCESS_DEFAULT);
+  test_trap_subprocess_with_logging_envp ("/logging/default-handler/subprocess/0x400",
+                                          NULL, NULL);
   g_test_trap_assert_passed ();
   g_test_trap_assert_stdout ("*LOG-0x400*message7*");
 
-  g_test_trap_subprocess ("/logging/default-handler/subprocess/would-drop", 0,
-                          G_TEST_SUBPROCESS_DEFAULT);
+  test_trap_subprocess_with_logging_envp ("/logging/default-handler/subprocess/would-drop",
+                                          NULL, NULL);
   g_test_trap_assert_passed ();
-  g_test_trap_subprocess ("/logging/default-handler/subprocess/would-drop-env1", 0,
-                          G_TEST_SUBPROCESS_DEFAULT);
+
+  test_trap_subprocess_with_logging_envp ("/logging/default-handler/subprocess/would-drop-env1",
+                                          NULL, NULL);
   g_test_trap_assert_passed ();
-  g_test_trap_subprocess ("/logging/default-handler/subprocess/would-drop-env2", 0,
-                          G_TEST_SUBPROCESS_DEFAULT);
+
+  test_trap_subprocess_with_logging_envp ("/logging/default-handler/subprocess/would-drop-env2",
+                                          "  bar    baz ", NULL);
   g_test_trap_assert_passed ();
-  g_test_trap_subprocess ("/logging/default-handler/subprocess/would-drop-env3", 0,
-                          G_TEST_SUBPROCESS_DEFAULT);
+
+  test_trap_subprocess_with_logging_envp ("/logging/default-handler/subprocess/would-drop-env3",
+                                          "foo bar", NULL);
   g_test_trap_assert_passed ();
-  g_test_trap_subprocess ("/logging/default-handler/subprocess/would-drop-env4", 0,
-                          G_TEST_SUBPROCESS_DEFAULT);
+
+  test_trap_subprocess_with_logging_envp ("/logging/default-handler/subprocess/would-drop-env4",
+                                          "all", NULL);
   g_test_trap_assert_passed ();
-  g_test_trap_subprocess ("/logging/default-handler/subprocess/would-drop-env5", 0,
-                          G_TEST_SUBPROCESS_DEFAULT);
+
+  test_trap_subprocess_with_logging_envp ("/logging/default-handler/subprocess/would-drop-env5",
+                                          "foobar", NULL);
   g_test_trap_assert_passed ();
-  g_test_trap_subprocess ("/logging/default-handler/subprocess/would-drop-robustness", 0,
-                          G_TEST_SUBPROCESS_DEFAULT);
+
+  test_trap_subprocess_with_logging_envp ("/logging/default-handler/subprocess/would-drop-env-systemd",
+                                          NULL, "1");
   g_test_trap_assert_passed ();
+
+  test_trap_subprocess_with_logging_envp ("/logging/default-handler/subprocess/would-drop-robustness",
+                                          NULL, NULL);
+  g_test_trap_assert_passed ();
+
+  test_trap_subprocess_with_logging_envp ("/logging/default-handler/subprocess/structured-logging-non-null-terminated-strings",
+                                          "foo", NULL);
+  g_test_trap_assert_passed ();
+  g_test_trap_assert_stdout_unmatched ("*bar*");
+  g_test_trap_assert_stdout_unmatched ("*bla*");
+  g_test_trap_assert_stdout ("*foo-DEBUG*baz*");
+}
+
+static void
+test_journald_handler (void)
+{
+  /* We can’t require that the journal exists on the test system. But if it
+   * does, we can check that the writer doesn’t crash. */
+  const GLogField fields[] = {
+    { "MESSAGE", "This is a test message.", -1 },
+    { "MESSAGE_ID", "7187d27ad7f84351b76b7612eef52cd6", -1 },
+    { "MY_APPLICATION_CUSTOM_FIELD", "some debug string", -1 },
+  };
+
+  g_log_writer_journald (G_LOG_LEVEL_DEBUG, fields, G_N_ELEMENTS (fields), NULL);
 }
 
 static void
@@ -543,15 +633,27 @@ test_fatal_log_mask (void)
   if (g_test_subprocess ())
     {
       g_log_set_fatal_mask ("bu", G_LOG_LEVEL_INFO);
+      g_assert_null (g_getenv ("G_MESSAGES_DEBUG"));
       g_log ("bu", G_LOG_LEVEL_INFO, "fatal");
       return;
     }
-  g_test_trap_subprocess (NULL, 0, G_TEST_SUBPROCESS_DEFAULT);
+  test_trap_subprocess_with_logging_envp (NULL, NULL, NULL);
   g_test_trap_assert_failed ();
   /* G_LOG_LEVEL_INFO isn't printed by default */
   g_test_trap_assert_stdout_unmatched ("*fatal*");
 }
 
+static void
+test_always_fatal (void)
+{
+  GLogLevelFlags log_level;
+
+  log_level = G_LOG_LEVEL_CRITICAL | G_LOG_LEVEL_WARNING;
+  g_log_set_always_fatal (log_level);
+
+  g_assert_cmpint (g_log_get_always_fatal (), ==, log_level | G_LOG_LEVEL_ERROR);
+}
+
 static gint my_print_count = 0;
 static void
 my_print_handler (const gchar *text)
@@ -1047,8 +1149,6 @@ test_structured_logging_set_writer_func_twice (void)
 int
 main (int argc, char *argv[])
 {
-  g_unsetenv ("G_MESSAGES_DEBUG");
-
   g_test_init (&argc, &argv, NULL);
 
   g_test_add_func ("/logging/default-handler", test_default_handler);
@@ -1073,9 +1173,13 @@ main (int argc, char *argv[])
   g_test_add_func ("/logging/default-handler/subprocess/would-drop-env3", test_default_handler_would_drop_env3);
   g_test_add_func ("/logging/default-handler/subprocess/would-drop-env4", test_default_handler_would_drop_env4);
   g_test_add_func ("/logging/default-handler/subprocess/would-drop-env5", test_default_handler_would_drop_env5);
+  g_test_add_func ("/logging/default-handler/subprocess/would-drop-env-systemd", test_default_handler_would_drop_env_systemd);
   g_test_add_func ("/logging/default-handler/subprocess/would-drop-robustness", test_default_handler_would_drop_robustness);
+  g_test_add_func ("/logging/default-handler/subprocess/structured-logging-non-null-terminated-strings", test_default_handler_structured_logging_non_nul_terminated_strings);
+  g_test_add_func ("/logging/journald-handler", test_journald_handler);
   g_test_add_func ("/logging/warnings", test_warnings);
   g_test_add_func ("/logging/fatal-log-mask", test_fatal_log_mask);
+  g_test_add_func ("/logging/always-fatal", test_always_fatal);
   g_test_add_func ("/logging/set-handler", test_set_handler);
   g_test_add_func ("/logging/print-handler", test_print_handler);
   g_test_add_func ("/logging/printerr-handler", test_printerr_handler);
